// -*- mode: c++; indent-tabs-mode: nil -*-
//! @file JsonRpcHandler.qm @brief JSON-RPC handler class definition for the HttpServer module

/*  JsonRpcHandler.qm Copyright (C) 2012 - 2016 David Nichols

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

// minimum qore version

// requires the binary json module

// need mime definitions

// need Util functions

// need HttpServerUtil definitions

// don't use "$" for vars, members, and method calls, assume local var scope


/** @mainpage JsonRpcHandler Module

    @tableofcontents

    @section jsonrpchandlerintro Introduction to the JsonRpcHandler Module

    This module implements server-side support for the <a href="http://en.wikipedia.org/wiki/JSON-RPC">JSON-RPC protocol</a> for serialization/deserialization of message data.

    This module provides the @ref JsonRpcHandler::JsonRpcHandler "JsonRpcHandler" class which can be used to provide an RPC handler for the HttpServer class provided by the HttpServer module.

    @section jsonrpchandlerexample Example Usage
    @code
%requires HttpServer
%requires JsonRpcHandler
%requires Mime

const ApiMethods = (
     ("name": "^sys\\.shutdown\$",
      "text": "sys.shutdown",
      "function": sub () { background http.stop();  return "OK"; },
      "help": "shuts down this server",
      "logopt": 0,
      ),
);

# a logging closure
code log = sub (string str) {printf("%y: %s\n", now_us(), str);};

# our bind address
const Bind = 8888;

JsonRpcHandler jsonRpcHandler(new AbstractAuthenticator(), api);
our HttpServer http(log, log);
http.addListener(Bind);
http.setHandler("jsonrpc", "", MimeTypeJsonRpc, jsonRpcHandler);
http.setDefaultHandler("jsonrpc", jsonRpcHandler);
log("now listening on %s\n", Bind);
http.waitStop();
    @endcode

    @section jsonrpchandler_relnotes JsonRpcHandler Release Notes

    @subsection jsonrpchandler_1_2 JsonRpcHandler v1.2
    - fixed bugs handling invalid JSON-RPC messages to give a user-friendly error response
    - converted to new-style

    @subsection jsonrpchandler_1_1 JsonRpcHandler v1.1
    - added an optional log closure to constructor

    @subsection jsonrpchandler_1_0 JsonRpcHandler v1.0
    - updated to a user module
*/

//! the %JsonRpcHandler namespace holds all public definitions in the %JsonRpcHandler module
namespace JsonRpcHandler {
    //! JsonRpcHandler class definition; to be registered as a handler in the HttpServer class
/**
     */
class JsonRpcHandler : public AbstractHttpRequestHandler {

public:
        //! implementation of the handler
        const Version = "1.0";

        //! internal methods of the handler (introspection)
        const InternalMethods = (
            ("function": "help",
             "help": "shows a list of JSON-RPC methods registered with this handler",
             "text": "help",
             "logopt": 2
            ),
            ("function": "listMethods",
             "help": "lists JSON-RPC method names registered with this handler",
             "text": "system.listMethods",
             "logopt": 2
            ),
            ("function": "system_describe",
             "help": "returns the service description object as per the JSON-RPC 1.1 spec",
             "text": "system.describe",
             "logopt": 2
            ));

        //! @cond nodoc
        private :
            list methods = ();
            hash mi;
            int loglevel;

            // if True then verbose exception info will be logged
            bool debug;

            // prefix to add to derived methods with GET requests if no "." characters are in the path
            __7_ string get_prefix;

            // a closure/call reference to get the log message and/or process arguments in incoming requests
            __7_ code getLogMessage;

            // a closure/call reference for logging (when set this is used instead of the HTTP server's logfunc for logging)
            __7_ code log_code;

public:
        //! @endcond

        //! creates the handler with the given method list
/** @param n_auth an authentication object (use new AbstractAuthenticator() for no authentication)
            @param n_methods a list of hashes with the following keys:
            - \c name: a regular expression to use for matching the method name
            - \c function: a string (giving a function name to call), a call reference, or a closure to call with the deserialized arguments to the method; the return value will be serialized to JSON-RPC and sent back to the caller
            - \c text: the human-readable name of the method
            - \c help: help text for the method
            - \c logopt: (optional - by convention) log options which can be used by a custom logger (see the getLogMessage parameter)
            - \c cmark: (optional) a data structure to be added to the context argument when the request method is called
            @param n_getLogMessage an optional closure or call reference to be called when an incoming request is received; if this is set then it will be called with the following arguments: a context hash (see HttpServer::AbstractHttpRequestHandler::handleRequest() for a description of the context hash), the method definition as passed in the methods argument to this constructor, and a reference to the arguments in this call
            @param n_dbg this parameter is set to @ref Qore::True "True", then additional information will be logged when errors occur
            @param n_get_prefix prefix to add to derived methods with GET requests if no "." characters are in the path
            @param n_log an optional closure or call reference to be called when logging

            @throw JSON-RPC-CONSTRUCTOR-ERROR missing \c "name", \c "function", or \"text" key in method hash, \c "function" key not assigned to a callable value
         */
        constructor(HttpServer::AbstractAuthenticator n_auth, list n_methods, __7_ code n_getLogMessage, bool n_dbg = False, __7_ string n_get_prefix, __7_ code n_log) ;


        //! adds a method to the handler dynamically
/** @param name a regular expression to use for matching the method name
            @param func a string (giving a function name to call), a call reference, or a closure to call with the deserialized arguments to the method; the return value will be serialized to XML-RPC and sent back to the caller
            @param text the human-readable name of the method
            @param help help text for the method
            @param logopt log options which can be used by a custom logger (see the getLogMessage parameter)
            @param cmark any value that will be added in the method hash under the \c "cmark" key
          */
        addMethod(string name, code func, string text, string help, any logopt, any cmark);


        //! turns on or off debugging; when debugging is enabled more verbose error messages are reported
        setDebug(bool dbg = True);


        //! returns the current status of the debug flag
        bool getDebug();


        //! @conf nodoc
        // don't reimplement this method; fix/enhance it in the module
        final private addMethodInternal(hash h);


        private hash help();


        private hash system_describe();


        private list listMethods();


        private log(hash cx, string str);


        // don't reimplement this method; fix/enhance it in the module
        final private hash callMethod(hash cx, any params);


        // method called by HttpServer to handle a request
        // don't reimplement this method; fix/enhance it in the module
        final hash handleRequest(hash cx, hash hdr, __7_ data body);

    };
};
